from __future__ import annotations

from abc import ABC, abstractmethod
from typing import (
    Any,
    Callable,
    Generic,
    Literal,
    Optional,
    Self,
    TypeVarTuple,
    Union,
    overload,
)

import numpy as np
from numpy.typing import NDArray
from scipy.optimize import newton
from typing_extensions import override

from relife import ParametricModel
from relife._plot import PlotParametricLifetimeModel
from relife.data import LifetimeData
from relife.likelihood import FittingResults, LikelihoodFromLifetimes
from relife.quadrature import (
    check_and_broadcast_bounds,
    legendre_quadrature,
    unweighted_laguerre_quadrature,
)

Args = TypeVarTuple("Args")


class ParametricLifetimeModel(ParametricModel, Generic[*Args], ABC):
    r"""Base class for lifetime model.

    This class defines the structure for creating lifetime model. It is a blueprint
    for implementing lifetime model expecting a variadic set of arguments.
    It expects implemantation of hazard functions (``hf``), cumulative hazard functions (``chf``),
    probability density function (``pdf``) and survival function (``sf``).
    Other functions are implemented by default but can be overridden by derived classes.

    Note:
        The abstract methods also provides a default implementation. One may not have to implement
        ``hf``, ``chf``, ``pdf`` and ``sf`` and just call ``super()`` to access the base implementation.

    Methods:
        hf: Abstract method to compute the hazard function.
        chf: Abstract method to compute the cumulative hazard function.
        sf: Abstract method to compute the survival function.
        pdf: Abstract method to compute the probability density function.

    Raises:
        NotImplementedError: Raised when an abstract method or feature in this
        class has not been implemented in a derived class.
    """

    fitting_results = Optional[FittingResults]  #: fitting results

    def __init__(self, **kwparams: Optional[float]):
        super().__init__(**kwparams)
        self.fitting_results = None

    @abstractmethod
    def sf(self, time: float | NDArray[np.float64], *args: *Args) -> np.float64 | NDArray[np.float64]:
        if hasattr(self, "chf"):
            return np.exp(
                -self.chf(
                    time,
                    *args,
                )
            )
        elif hasattr(self, "pdf") and hasattr(self, "hf"):
            return self.pdf(time, *args) / self.hf(time, *args)
        else:
            class_name = type(self).__name__
            raise NotImplementedError(
                f"""
                {class_name} must implement concrete sf function
                """
            )

    @abstractmethod
    def hf(self, time: float | NDArray[np.float64], *args: *Args) -> np.float64 | NDArray[np.float64]:
        if hasattr(self, "pdf") and hasattr(self, "sf"):
            return self.pdf(time, *args) / self.sf(time, *args)
        else:
            class_name = type(self).__name__
            raise NotImplementedError(
                f"""
                    {class_name} must implement concrete hf function
                    """
            )

    @abstractmethod
    def chf(self, time: float | NDArray[np.float64], *args: *Args) -> np.float64 | NDArray[np.float64]:
        if hasattr(self, "sf"):
            return -np.log(self.sf(time, *args))
        elif hasattr(self, "pdf") and hasattr(self, "hf"):
            return -np.log(self.pdf(time, *args) / self.hf(time, *args))
        else:
            class_name = type(self).__name__
            raise NotImplementedError(
                f"""
                {class_name} must implement concrete chf or at least concrete hf function
                """
            )

    @abstractmethod
    def pdf(self, time: float | NDArray[np.float64], *args: *Args) -> np.float64 | NDArray[np.float64]:
        try:
            return self.sf(time, *args) * self.hf(time, *args)
        except NotImplementedError as err:
            class_name = type(self).__name__
            raise NotImplementedError(
                f"""
            {class_name} must implement pdf or the above functions
            """
            ) from err

    def cdf(self, time: float | NDArray[np.float64], *args: *Args) -> np.float64 | NDArray[np.float64]:
        return 1 - self.sf(time, *args)

    def ppf(
        self,
        probability: float | NDArray[np.float64],
        *args: *Args,
    ) -> np.float64 | NDArray[np.float64]:
        return self.isf(1 - probability, *args)

    def median(self, *args: *Args) -> np.float64 | NDArray[np.float64]:
        return self.ppf(0.5, *args)

    def isf(
        self,
        probability: float | NDArray[np.float64],
        *args: *Args,
    ) -> np.float64 | NDArray[np.float64]:
        # return type of func cannot be NDArray, thus np.sum is needed (see scipy type stubs)
        func: Callable[[float | NDArray[np.float64]], np.float64] = lambda x: self.sf(x, *args) - probability
        return newton(
            func,
            x0=np.zeros_like(probability),
            args=args,
        )

    def ichf(
        self,
        cumulative_hazard_rate: float | NDArray[np.float64],
        *args: *Args,
    ) -> np.float64 | NDArray[np.float64]:

        # return type of func cannot be NDArray, thus np.sum is needed (see scipy type stubs)
        func: Callable[[float | NDArray[np.float64]], np.float64] = lambda x: np.sum(
            self.chf(x, *args) - cumulative_hazard_rate
        )
        return newton(
            func,
            x0=np.zeros_like(cumulative_hazard_rate),
        )

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        *args: *Args,
        return_event: Literal[False],
        return_entry: Literal[False],
        seed: Optional[int] = None,
    ) -> np.float64 | NDArray[np.float64]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        *args: *Args,
        return_event: Literal[True],
        return_entry: Literal[False],
        seed: Optional[int] = None,
    ) -> tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_]]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        *args: *Args,
        return_event: Literal[False],
        return_entry: Literal[True],
        seed: Optional[int] = None,
    ) -> tuple[np.float64 | NDArray[np.float64], np.float64 | NDArray[np.float64]]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        *args: *Args,
        return_event: Literal[True],
        return_entry: Literal[True],
        seed: Optional[int] = None,
    ) -> tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_], np.float64 | NDArray[np.float64]]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        *args: *Args,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Optional[int] = None,
    ) -> Union[
        np.float64 | NDArray[np.float64],
        tuple[np.float64 | NDArray[np.float64], np.float64 | NDArray[np.float64]],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_]],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_], np.float64 | NDArray[np.float64]],
    ]: ...

    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        *args: *Args,
        return_event: bool = False,
        return_entry: bool = False,
        seed: Optional[int] = None,
    ) -> Union[
        np.float64 | NDArray[np.float64],
        tuple[np.float64 | NDArray[np.float64], np.float64 | NDArray[np.float64]],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_]],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_], np.float64 | NDArray[np.float64]],
    ]:
        rs = np.random.RandomState(seed=seed)
        probability = rs.uniform(size=size)
        time = self.isf(probability, *args)
        event = np.ones_like(time, dtype=np.bool_) if isinstance(time, np.ndarray) else np.bool_(True)
        entry = np.zeros_like(time, dtype=np.float64) if isinstance(time, np.ndarray) else np.float64(0)
        if not return_event and not return_entry:
            return time
        elif return_event and not return_entry:
            return time, event
        elif not return_event and return_entry:
            return time, entry
        else:
            return time, event, entry

    @property
    def plot(self) -> PlotParametricLifetimeModel:
        """Get plot"""
        return PlotParametricLifetimeModel(self)

    def ls_integrate(
        self,
        func: Callable[[float | NDArray[np.float64]], np.float64 | NDArray[np.float64]],
        a: float | NDArray[np.float64],
        b: float | NDArray[np.float64],
        *args: *Args,
        deg: int = 10,
    ) -> np.float64 | NDArray[np.float64]:

        def integrand(x: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
            #  x.shape == (deg,), (deg, n) or (deg, m, n), ie points of quadratures
            # fx : (d_1, ..., d_i, deg), (d_1, ..., d_i, deg, n) or (d_1, ..., d_i, deg, m, n)
            x = np.asarray(x)
            fx = func(x)
            if fx.shape[-len(x.shape) :] != x.shape:
                raise ValueError(
                    f"""
                    func can't squeeze input dimensions. If x has shape (d_1, ..., d_i), func(x) must have shape (..., d_1, ..., d_i).
                    Ex : if x.shape == (m, n), func(x).shape == (..., m, n).
                    """
                )
            if x.ndim == 3:  # reshape because model.pdf is tested only for input ndim <= 2
                xdeg, m, n = x.shape
                x = np.rollaxis(x, 1).reshape(m, -1)  # (m, deg*n), roll on m because axis 0 must align with m of args
                pdf = self.pdf(x, *args)  # (m, deg*n)
                pdf = np.rollaxis(pdf.reshape(m, xdeg, n), 1, 0)  #  (deg, m, n)
            else:  # ndim == 1 | 2
                # reshape to (1, deg*n) or (1, deg), ie place 1 on axis 0 to allow broadcasting with m of args
                pdf = self.pdf(x.reshape(1, -1), *args)  # (1, deg*n) or (1, deg)
                pdf = pdf.reshape(x.shape)  # (deg, n) or (deg,)

            # (d_1, ..., d_i, deg) or (d_1, ..., d_i, deg, n) or (d_1, ..., d_i, deg, m, n)
            return fx * pdf

        arr_a, arr_b = check_and_broadcast_bounds(a, b)  # (), (n,) or (m, n)
        if np.any(arr_a > arr_b):
            raise ValueError("Bound values a must be lower than values of b")

        bound_b = self.isf(1e-4, *args)  #  () or (m, 1), if (m, 1) then arr_b.shape == (m, 1) or (m, n)
        broadcasted_arrs = np.broadcast_arrays(arr_a, arr_b, bound_b)
        arr_a = broadcasted_arrs[0].copy()  # arr_a.shape == arr_b.shape == bound_b.shape
        arr_b = broadcasted_arrs[1].copy()  # arr_a.shape == arr_b.shape == bound_b.shape
        bound_b = broadcasted_arrs[2].copy()  # arr_a.shape == arr_b.shape == bound_b.shape
        is_inf = np.isinf(arr_b)  # () or (n,) or (m, n)
        arr_b = np.where(is_inf, bound_b, arr_b)
        integration = legendre_quadrature(
            integrand, arr_a, arr_b, deg=deg
        )  #  (d_1, ..., d_i), (d_1, ..., d_i, n) or (d_1, ..., d_i, m, n)
        is_inf, _ = np.broadcast_arrays(is_inf, integration)
        return np.where(
            is_inf,
            integration + unweighted_laguerre_quadrature(integrand, arr_b, deg=deg),
            integration,
        )

    @overload
    def moment(self, n: int) -> np.float64: ...

    @overload
    def moment(self, n: int, *args: *Args) -> np.float64 | NDArray[np.float64]: ...

    def moment(self, n: int, *args: *Args) -> np.float64 | NDArray[np.float64]:
        if n < 1:
            raise ValueError("order of the moment must be at least 1")
        func: Callable[[float | NDArray[np.float64]], np.float64 | NDArray[np.float64]] = lambda x: np.power(x, n)
        return self.ls_integrate(
            func,
            0.0,
            np.inf,
            *args,
            deg=100,
        )  #  high degree of polynome to ensure high precision

    def mean(self, *args: *Args) -> np.float64 | NDArray[np.float64]:
        return self.moment(1, *args)

    def var(self, *args: *Args) -> np.float64 | NDArray[np.float64]:
        return self.moment(2, *args) - self.moment(1, *args) ** 2

    def mrl(self, time: float | NDArray[np.float64], *args: *Args) -> np.float64 | NDArray[np.float64]:
        sf = self.sf(time, *args)
        func: Callable[[float | NDArray[np.float64]], NDArray[np.float64]] = lambda x: np.asarray(x) - time
        ls = self.ls_integrate(func, time, np.inf, *args)
        if sf.ndim < 2:  # 2d to 1d or 0d
            ls = np.squeeze(ls)
        return ls / sf


class LifetimeDistribution(ParametricLifetimeModel[()], ABC):
    """
    Base class for distribution model.
    """

    @override
    def sf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return super().sf(time)

    @override
    def isf(self, probability: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        cumulative_hazard_rate = -np.log(probability + 1e-6)  # avoid division by zero
        return self.ichf(cumulative_hazard_rate)

    @override
    def cdf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return super().cdf(time)

    def pdf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return super().pdf(time)

    @overload
    def ppf(self, probability: float) -> np.float64: ...

    @overload
    def ppf(self, probability: NDArray[np.float64]) -> NDArray[np.float64]: ...

    @overload
    def ppf(self, probability: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]: ...

    @override
    def ppf(self, probability: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return super().ppf(probability)

    @override
    def median(self) -> np.float64:
        return self.ppf(0.5)  # no super here to return np.float64

    @abstractmethod
    def dhf(
        self,
        time: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]: ...

    @overload
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[False],
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[True],
    ) -> np.float64 | NDArray[np.float64]: ...

    @abstractmethod
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[False],
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[True],
    ) -> np.float64 | NDArray[np.float64]: ...

    @abstractmethod
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_sf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[False],
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_sf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[True],
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_sf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64 | NDArray[np.float64], ...]:
        jac_chf, sf = self.jac_chf(time, asarray=True), self.sf(time)
        jac = -jac_chf * sf
        if not asarray:
            return np.unstack(jac)
        return jac

    @overload
    def jac_cdf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[False] = False,
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_cdf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[True] = True,
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_cdf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64 | NDArray[np.float64], ...]:
        jac = -self.jac_sf(time, asarray=True)
        if not asarray:
            return np.unstack(jac)
        return jac

    @overload
    def jac_pdf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[False] = False,
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_pdf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: Literal[True] = True,
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_pdf(
        self,
        time: float | NDArray[np.float64],
        *,
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64 | NDArray[np.float64], ...]:
        jac_hf, hf = self.jac_hf(time, asarray=True), self.hf(time)
        jac_sf, sf = self.jac_sf(time, asarray=True), self.sf(time)
        jac = jac_hf * sf + jac_sf * hf
        if not asarray:
            return np.unstack(jac)
        return jac

    @override
    def moment(self, n: int) -> np.float64:
        return super().moment(n)

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        return_event: Literal[False],
        return_entry: Literal[False],
        seed: Optional[int] = None,
    ) -> np.float64 | NDArray[np.float64]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        return_event: Literal[True],
        return_entry: Literal[False],
        seed: Optional[int] = None,
    ) -> tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_]]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        return_event: Literal[False],
        return_entry: Literal[True],
        seed: Optional[int] = None,
    ) -> tuple[np.float64 | NDArray[np.float64], np.float64 | NDArray[np.float64]]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        return_event: Literal[True],
        return_entry: Literal[True],
        seed: Optional[int] = None,
    ) -> tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_], np.float64 | NDArray[np.float64]]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        return_event: bool = False,
        return_entry: bool = False,
        seed: Optional[int] = None,
    ) -> Union[
        np.float64 | NDArray[np.float64],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_]],
        tuple[np.float64 | NDArray[np.float64], np.float64 | NDArray[np.float64]],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_], np.float64 | NDArray[np.float64]],
    ]: ...

    @override
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        return_event: bool = False,
        return_entry: bool = False,
        seed: Optional[int] = None,
    ) -> Union[
        np.float64 | NDArray[np.float64],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_]],
        tuple[np.float64 | NDArray[np.float64], np.float64 | NDArray[np.float64]],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_], np.float64 | NDArray[np.float64]],
    ]:
        return super().rvs(size, return_event=return_event, return_entry=return_entry, seed=seed)

    @override
    def ls_integrate(
        self,
        func: Callable[[float | NDArray[np.float64]], np.float64 | NDArray[np.float64]],
        a: float | NDArray[np.float64],
        b: float | NDArray[np.float64],
        deg: int = 10,
    ) -> np.float64 | NDArray[np.float64]:

        return super().ls_integrate(func, a, b, deg=deg)

    def fit(
        self,
        time: NDArray[np.float64],
        event: Optional[NDArray[np.bool_]] = None,
        entry: Optional[NDArray[np.float64]] = None,
        departure: Optional[NDArray[np.float64]] = None,
        **kwargs: Any,
    ) -> Self:
        # initialize params structure (number of parameters in params tree)
        lifetime_data: LifetimeData = LifetimeData(time, event=event, entry=entry, departure=departure)
        likelihood = LikelihoodFromLifetimes(self, lifetime_data)
        fitting_results = likelihood.maximum_likelihood_estimation(**kwargs)
        self.params = fitting_results.optimal_params
        self.fitting_results = fitting_results
        return self


class CovarEffect(ParametricModel):
    """
    Covariates effect.

    Parameters
    ----------
    *coefficients : float
        Coefficients of the covariates effect.
    """

    def __init__(self, coefficients: tuple[Optional[float], ...] = (None,)):
        super().__init__(**{f"coef_{i + 1}": v for i, v in enumerate(coefficients)})

    @property
    def nb_coef(self) -> int:
        return self.nb_params

    def g(self, covar: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        """
        Compute the covariates effect.
        If covar.shape : () or (nb_coef,) => out.shape : (), float
        If covar.shape : (m, nb_coef) => out.shape : (m, 1)
        """
        arr_covar: NDArray[np.float64] = np.asarray(covar)  # (), (nb_coef,) or (m, nb_coef)
        if arr_covar.ndim > 2:
            raise ValueError(f"Invalid covar shape. Expected (nb_coef,) or (m, nb_coef) but got {arr_covar.shape}")
        covar_nb_coef = arr_covar.size if arr_covar.ndim <= 1 else arr_covar.shape[-1]
        if covar_nb_coef != self.nb_coef:
            raise ValueError(
                f"Invalid covar. Number of covar does not match number of coefficients. Got {self.nb_coef} nb_coef but covar shape is {arr_covar.shape}"
            )
        g = np.exp(np.sum(self.params * arr_covar, axis=-1, keepdims=True))  # (m, 1)
        if arr_covar.ndim <= 1:
            return np.float64(g.item())
        return g

    def jac_g(
        self, covar: float | NDArray[np.float64], *, asarray: bool = False
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64, ...] | tuple[NDArray[np.float64], ...]:
        """
        Compute the Jacobian of the covariates effect.
        If covar.shape : () or (nb_coef,) => out.shape : (nb_coef,)
        If covar.shape : (m, nb_coef) => out.shape : (nb_coef, m, 1)
        """
        arr_covar: NDArray[np.float64] = np.asarray(covar)  # (), (nb_coef,) or (m, nb_coef)
        g = self.g(arr_covar)  # () or (m, 1)
        jac = arr_covar.T.reshape(self.nb_coef, -1, 1) * g  # (nb_coef, m, 1)
        if arr_covar.ndim <= 1:
            jac = jac.reshape(self.nb_coef)  # (nb_coef,) or (nb_coef, m, 1)
        if not asarray:
            return np.unstack(jac, axis=0)  # tuple
        return jac  # (nb_coef, m, 1)


class LifetimeRegression(
    ParametricLifetimeModel[float | NDArray[np.float64], *tuple[float | NDArray[np.float64], ...]], ABC
):
    """
    Base class for regression model.

    At least one positional covar arg and 0 or more additional args (variable number)
    see : https://peps.python.org/pep-0646/#unpacking-unbounded-tuple-types*

    Note:
    LifetimeRegression does not preserve generic : at the moment, additional args are supposed to be always float | NDArray[np.float64]
    """

    def __init__(
        self,
        baseline: LifetimeDistribution | LifetimeRegression,
        coefficients: tuple[Optional[float], ...] = (None,),
    ):
        super().__init__()
        self.covar_effect = CovarEffect(coefficients)
        self.baseline = baseline

    @property
    def coefficients(self) -> NDArray[np.float64]:
        """Get the coefficients values of the covariate effect.

        Returns
        -------
        ndarray
        """
        return self.covar_effect.params

    @override
    def sf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        return super().sf(time, covar, *args)

    @override
    def isf(
        self,
        probability: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        cumulative_hazard_rate = -np.log(probability + 1e-6)  # avoid division by zero
        return self.ichf(cumulative_hazard_rate, covar, *args)

    @override
    def cdf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        return super().cdf(time, *(covar, *args))

    def pdf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        return super().pdf(time, *(covar, *args))

    @override
    def ppf(
        self,
        probability: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        return super().ppf(probability, *(covar, *args))

    @override
    def mrl(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        return super().mrl(time, *(covar, *args))

    @override
    def ls_integrate(
        self,
        func: Callable[[float | NDArray[np.float64]], np.float64 | NDArray[np.float64]],
        a: float | NDArray[np.float64],
        b: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        deg: int = 10,
    ) -> np.float64 | NDArray[np.float64]:
        return super().ls_integrate(func, a, b, *(covar, *args), deg=deg)

    @override
    def mean(
        self, covar: float | NDArray[np.float64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        return super().mean(*(covar, *args))

    @override
    def var(
        self, covar: float | NDArray[np.float64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        return super().var(*(covar, *args))

    @override
    def median(
        self, covar: float | NDArray[np.float64], *args: float | NDArray[np.float64]
    ) -> np.float64 | NDArray[np.float64]:
        return super().median(*(covar, *args))

    @abstractmethod
    def dhf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]: ...

    @overload
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: Literal[False] = False,
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: Literal[True] = True,
    ) -> np.float64 | NDArray[np.float64]: ...

    @abstractmethod
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: Literal[False] = False,
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: Literal[True] = True,
    ) -> np.float64 | NDArray[np.float64]: ...

    @abstractmethod
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_sf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: Literal[False] = False,
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_sf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: Literal[True] = True,
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_sf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64 | NDArray[np.float64], ...]:
        jac = -self.jac_chf(time, covar, *args, asarray=True) * self.sf(time, covar, *args)
        if not asarray:
            return np.unstack(jac)
        return jac

    @overload
    def jac_cdf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: Literal[False],
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_cdf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: Literal[True],
    ) -> np.float64 | NDArray[np.float64]: ...

    @overload
    def jac_cdf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: bool,
    ) -> tuple[np.float64 | NDArray[np.float64], ...] | np.float64 | NDArray[np.float64]: ...

    def jac_cdf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64 | NDArray[np.float64], ...]:
        jac = -self.jac_sf(time, covar, *args, asarray=True)
        if not asarray:
            return np.unstack(jac)
        return jac

    @overload
    def jac_pdf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: Literal[False] = False,
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_pdf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: Literal[True] = True,
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_pdf(
        self,
        time: float | NDArray[np.float64],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64 | NDArray[np.float64], ...]:
        jac = self.jac_hf(time, covar, *args, asarray=True) * self.sf(time, covar, *args) + self.jac_sf(
            time, covar, *args, asarray=True
        ) * self.hf(time, covar, *args)
        if not asarray:
            return np.unstack(jac)
        return jac

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        return_event: Literal[False],
        return_entry: Literal[False],
        seed: Optional[int] = None,
    ) -> np.float64 | NDArray[np.float64]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        return_event: Literal[True],
        return_entry: Literal[False],
        seed: Optional[int] = None,
    ) -> tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_]]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        return_event: Literal[False],
        return_entry: Literal[True],
        seed: Optional[int] = None,
    ) -> tuple[np.float64 | NDArray[np.float64], np.float64 | NDArray[np.float64]]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        return_event: Literal[True],
        return_entry: Literal[True],
        seed: Optional[int] = None,
    ) -> tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_], np.float64 | NDArray[np.float64]]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        return_event: bool = False,
        return_entry: bool = False,
        seed: Optional[int] = None,
    ) -> Union[
        np.float64 | NDArray[np.float64],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_]],
        tuple[np.float64 | NDArray[np.float64], np.float64 | NDArray[np.float64]],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_], np.float64 | NDArray[np.float64]],
    ]: ...

    @override
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
        return_event: bool = False,
        return_entry: bool = False,
        seed: Optional[int] = None,
    ) -> Union[
        np.float64 | NDArray[np.float64],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_]],
        tuple[np.float64 | NDArray[np.float64], np.float64 | NDArray[np.float64]],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_], np.float64 | NDArray[np.float64]],
    ]:
        return super().rvs(size, *(covar, *args), return_event=return_event, return_entry=return_entry, seed=seed)

    def _init_coefficients(self, covar: float | NDArray[np.float64], *args: float | NDArray[np.float64]) -> None:
        covar = np.atleast_2d(np.asarray(covar, dtype=np.float64))
        self.covar_effect = CovarEffect((None,) * covar.shape[-1])  # set new covar_effet -> changes params structure
        if hasattr(self.baseline, "_init_coefficients"):  # recursion in case of PPH(AFT(...))
            self.baseline._init_coefficients(*args)

    def fit(
        self,
        time: NDArray[np.float64],
        covar: NDArray[np.float64],
        *args: NDArray[np.float64],
        event: Optional[NDArray[np.bool_]] = None,
        entry: Optional[NDArray[np.float64]] = None,
        departure: Optional[NDArray[np.float64]] = None,
        **kwargs: Any,
    ) -> Self:
        # change number of coefficients depending on covar passed
        self._init_coefficients(covar, *args)
        lifetime_data: LifetimeData = LifetimeData(
            time, event=event, entry=entry, departure=departure, args=(covar, *args)
        )
        likelihood = LikelihoodFromLifetimes(self, lifetime_data)
        fitting_results = likelihood.maximum_likelihood_estimation(**kwargs)
        self.params = fitting_results.optimal_params
        self.fitting_results = fitting_results
        return self


class NonParametricLifetimeModel(ABC):

    @abstractmethod
    def fit(
        self,
        time: NDArray[np.float64],
        event: Optional[NDArray[np.bool_]] = None,
        entry: Optional[NDArray[np.float64]] = None,
        departure: Optional[NDArray[np.float64]] = None,
    ) -> Self: ...


class FrozenParametricLifetimeModel(ParametricModel, Generic[*Args]):

    frozen_args: tuple[*Args]

    def __init__(self, model: ParametricLifetimeModel[*Args], args_nb_assets: int, *args: *Args):
        super().__init__()
        if np.any(np.isnan(model.params)):
            raise ValueError("You try to freeze a model with unsetted parameters. Set params first")
        self.unfrozen_model = model
        self.frozen_args = args
        self.args_nb_assets = args_nb_assets

    def unfreeze(self) -> ParametricLifetimeModel[*Args]:
        return self.unfrozen_model

    def hf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return self.unfrozen_model.hf(time, *self.frozen_args)

    def chf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return self.unfrozen_model.chf(time, *self.frozen_args)

    def sf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return self.unfrozen_model.sf(time, *self.frozen_args)

    def pdf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return self.unfrozen_model.pdf(time, *self.frozen_args)

    def mrl(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return self.unfrozen_model.mrl(time, *self.frozen_args)

    def moment(self, n: int) -> np.float64 | NDArray[np.float64]:
        return self.unfrozen_model.moment(n, *self.frozen_args)

    def mean(self) -> np.float64 | NDArray[np.float64]:
        return self.unfrozen_model.moment(1, *self.frozen_args)

    def var(self) -> np.float64 | NDArray[np.float64]:
        return self.unfrozen_model.moment(2, *self.frozen_args) - self.unfrozen_model.moment(1, *self.frozen_args) ** 2

    def isf(self, probability: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return self.unfrozen_model.isf(probability, *self.frozen_args)

    def ichf(self, cumulative_hazard_rate: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return self.unfrozen_model.ichf(cumulative_hazard_rate, *self.frozen_args)

    def cdf(self, time: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return self.unfrozen_model.cdf(time, *self.frozen_args)

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        return_event: Literal[False],
        return_entry: Literal[False],
        seed: Optional[int] = None,
    ) -> np.float64 | NDArray[np.float64]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        return_event: Literal[True],
        return_entry: Literal[False],
        seed: Optional[int] = None,
    ) -> tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_]]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        return_event: Literal[False],
        return_entry: Literal[True],
        seed: Optional[int] = None,
    ) -> tuple[np.float64 | NDArray[np.float64], np.float64 | NDArray[np.float64]]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        return_event: Literal[True],
        return_entry: Literal[True],
        seed: Optional[int] = None,
    ) -> tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_], np.float64 | NDArray[np.float64]]: ...

    @overload
    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        return_event: bool = False,
        return_entry: bool = False,
        seed: Optional[int] = None,
    ) -> Union[
        np.float64 | NDArray[np.float64],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_]],
        tuple[np.float64 | NDArray[np.float64], np.float64 | NDArray[np.float64]],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_], np.float64 | NDArray[np.float64]],
    ]: ...

    def rvs(
        self,
        size: int | tuple[int] | tuple[int, int],
        return_event: bool = False,
        return_entry: bool = False,
        seed: Optional[int] = None,
    ) -> Union[
        np.float64 | NDArray[np.float64],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_]],
        tuple[np.float64 | NDArray[np.float64], np.float64 | NDArray[np.float64]],
        tuple[np.float64 | NDArray[np.float64], np.bool_ | NDArray[np.bool_], np.float64 | NDArray[np.float64]],
    ]:
        return self.unfrozen_model.rvs(
            size, *self.frozen_args, return_event=return_event, return_entry=return_entry, seed=seed
        )

    def ppf(self, probability: float | NDArray[np.float64]) -> np.float64 | NDArray[np.float64]:
        return self.unfrozen_model.ppf(probability, *self.frozen_args)

    def median(self) -> np.float64 | NDArray[np.float64]:
        return self.unfrozen_model.median(*self.frozen_args)

    def ls_integrate(
        self,
        func: Callable[[float | NDArray[np.float64]], np.float64 | NDArray[np.float64]],
        a: float | NDArray[np.float64],
        b: float | NDArray[np.float64],
        deg: int = 10,
    ) -> np.float64 | NDArray[np.float64]:
        return self.unfrozen_model.ls_integrate(func, a, b, *self.frozen_args, deg=deg)


class FrozenLifetimeRegression(
    FrozenParametricLifetimeModel[float | NDArray[np.float64], *tuple[float | NDArray[np.float64], ...]]
):
    r"""
    Frozen lifetime regression.

    Parameters
    ----------
    model : LifetimeRegression
        Any lifetime regression.
    args_nb_assets : int
        Number of assets given in frozen arguments. It is automatically computed by ``freeze`` function.
    covar : float or np.ndarray
        Covariate values to be frozen.
    *args : float or np.ndarray
        Additional arguments needed by the model to be frozen.


    Attributes
    ----------
    unfrozen_model : LifetimeRegression
        The unfrozen regression model.
    frozen_args : tuple of float or np.ndarray
        All the frozen arguments given and necessary to compute model functions.
    args_nb_assets : int
        Number of assets passed in frozen arguments. The data is mainly used to control numpy broadcasting and may not
        interest an user.

    Warnings
    --------
    This class is documented for the purpose of clarity and mainly address contributors or advance users. Actually, the
    recommanded way to instanciate a ``FrozenLifetimeRegression`` is use to ``freeze`` factory function.

    """

    unfrozen_model: LifetimeRegression
    frozen_args: tuple[float | NDArray[np.float64], *tuple[float | NDArray[np.float64], ...]]

    @override
    def __init__(
        self,
        model: LifetimeRegression,
        args_nb_assets: int,
        covar: float | NDArray[np.float64],
        *args: float | NDArray[np.float64],
    ):
        super().__init__(model, args_nb_assets, *(covar, *args))

    @override
    def unfreeze(self) -> LifetimeRegression:
        return self.unfrozen_model

    @property
    def nb_coef(self) -> int:
        return self.unfrozen_model.nb_coef

    @property
    def covar(self) -> float | NDArray[np.float64]:
        return self.frozen_args[0]

    @covar.setter
    def covar(self, value: float | NDArray[np.float64]) -> None:
        self.frozen_args = (value,) + self.frozen_args[1:]

    def dhf(
        self,
        time: float | NDArray[np.float64],
    ) -> np.float64 | NDArray[np.float64]:
        return self.unfrozen_model.dhf(time, self.frozen_args[0], *self.frozen_args[1:])

    @overload
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        asarray: Literal[False] = False,
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        asarray: Literal[True] = True,
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_hf(
        self,
        time: float | NDArray[np.float64],
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64, ...] | tuple[NDArray[np.float64], ...]:
        return self.unfrozen_model.jac_hf(time, self.frozen_args[0], *self.frozen_args[1:], asarray=asarray)

    @overload
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        asarray: Literal[False] = False,
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        asarray: Literal[True] = True,
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_chf(
        self,
        time: float | NDArray[np.float64],
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64, ...] | tuple[NDArray[np.float64], ...]:
        return self.unfrozen_model.jac_chf(time, self.frozen_args[0], *self.frozen_args[1:], asarray=asarray)

    @overload
    def jac_sf(
        self,
        time: float | NDArray[np.float64],
        asarray: Literal[False] = False,
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_sf(
        self,
        time: float | NDArray[np.float64],
        asarray: Literal[True] = True,
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_sf(
        self,
        time: float | NDArray[np.float64],
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64, ...] | tuple[NDArray[np.float64], ...]:
        return self.unfrozen_model.jac_sf(time, self.frozen_args[0], *self.frozen_args[1:], asarray=asarray)

    @overload
    def jac_cdf(
        self,
        time: float | NDArray[np.float64],
        asarray: Literal[False],
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_cdf(
        self,
        time: float | NDArray[np.float64],
        asarray: Literal[True],
    ) -> np.float64 | NDArray[np.float64]: ...

    @overload
    def jac_cdf(
        self,
        time: float | NDArray[np.float64],
        asarray: bool,
    ) -> tuple[np.float64 | NDArray[np.float64], ...] | np.float64 | NDArray[np.float64]: ...

    def jac_cdf(
        self,
        time: float | NDArray[np.float64],
        asarray: bool = False,
    ) -> tuple[np.float64 | NDArray[np.float64], ...] | np.float64 | NDArray[np.float64]:
        return self.unfrozen_model.jac_cdf(time, self.frozen_args[0], *self.frozen_args[1:], asarray=asarray)

    @overload
    def jac_pdf(
        self,
        time: float | NDArray[np.float64],
        asarray: Literal[False] = False,
    ) -> tuple[np.float64 | NDArray[np.float64], ...]: ...

    @overload
    def jac_pdf(
        self,
        time: float | NDArray[np.float64],
        asarray: Literal[True] = True,
    ) -> np.float64 | NDArray[np.float64]: ...

    def jac_pdf(
        self,
        time: float | NDArray[np.float64],
        asarray: bool = False,
    ) -> np.float64 | NDArray[np.float64] | tuple[np.float64, ...] | tuple[NDArray[np.float64], ...]:
        return self.unfrozen_model.jac_pdf(time, self.frozen_args[0], *self.frozen_args[1:], asarray=asarray)


TIME_BASE_DOCSTRING = """
{name}.

Parameters
----------
time : float or np.ndarray
    Elapsed time value(s) at which to compute the function.
    If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

Returns
-------
np.float64 or np.ndarray
    Function values at each given time(s).
"""

JAC_BASE_DOCSTRING = """
{name}.

Parameters
----------
time : float or np.ndarray
    Elapsed time value(s) at which to compute the function.
    If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.
asarray : bool, default is False

Returns
-------
np.float64, np.ndarray or tuple of np.float64 or np.ndarray
    The derivatives with respect to each parameter. If ``asarray`` is False, the function returns a tuple containing
    the same number of elements as parameters. If ``asarray`` is True, the function returns an ndarray
    whose first dimension equals the number of parameters. This output is equivalent to applying ``np.stack`` on the output
    tuple when ``asarray`` is False.
"""

MOMENT_BASE_DOCSTRING = """
{name}.

Returns
-------
np.float64
    {name} value.
"""

PROBABILITY_BASE_DOCSTRING = """
{name}.

Parameters
----------
probability : float or np.ndarray
    Probability value(s) at which to compute the function.
    If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

Returns
-------
np.float64 or np.ndarray
    Function values at each given probability value(s).
"""

FrozenParametricLifetimeModel.sf.__doc__ = TIME_BASE_DOCSTRING.format(name="The survival function")
FrozenParametricLifetimeModel.hf.__doc__ = TIME_BASE_DOCSTRING.format(name="The hazard function")
FrozenParametricLifetimeModel.chf.__doc__ = TIME_BASE_DOCSTRING.format(name="The cumulative hazard function")
FrozenParametricLifetimeModel.pdf.__doc__ = TIME_BASE_DOCSTRING.format(name="The probability density function")
FrozenParametricLifetimeModel.cdf.__doc__ = TIME_BASE_DOCSTRING.format(name="The cumulative distribution function")
FrozenParametricLifetimeModel.mrl.__doc__ = TIME_BASE_DOCSTRING.format(name="The mean residual life function")
FrozenParametricLifetimeModel.ppf.__doc__ = PROBABILITY_BASE_DOCSTRING.format(name="The percent point function")
FrozenParametricLifetimeModel.ppf.__doc__ += f"""
Notes
-----
The ``ppf`` is the inverse of :py:meth:`~FrozenParametricLifetimeModel.cdf`.
"""
FrozenParametricLifetimeModel.isf.__doc__ = PROBABILITY_BASE_DOCSTRING.format(name="Inverse survival function")

FrozenParametricLifetimeModel.rvs.__doc__ = """
Random variable sampling.

Parameters
----------
size : int, (int,) or (int, int)
    Size of the generated sample. If size is ``n`` or ``(n,)``, n samples are generated. If size is ``(m,n)``, a 
    2d array of samples is generated. 
return_event : bool, default is False
    If True, returns event indicators along with the sample time values.
random_entry : bool, default is False
    If True, returns corresponding entry values of the sample time values.
seed : optional int, default is None
    Random seed used to fix random sampling.

Returns
-------
float, ndarray or tuple of float or ndarray
    The sample values. If either ``return_event`` or ``random_entry`` is True, returns a tuple containing
    the time values followed by event values, entry values or both.
"""

FrozenParametricLifetimeModel.ls_integrate.__doc__ = """
Lebesgue-Stieltjes integration.

Parameters
----------
func : callable (in : 1 ndarray , out : 1 ndarray)
    The callable must have only one ndarray object as argument and one ndarray object as output
a : ndarray (maximum number of dimension is 2)
    Lower bound(s) of integration.
b : ndarray (maximum number of dimension is 2)
    Upper bound(s) of integration. If lower bound(s) is infinite, use np.inf as value.)
deg : int, default 10
    Degree of the polynomials interpolation

Returns
-------
np.ndarray
    Lebesgue-Stieltjes integral of func from `a` to `b`.
"""

FrozenParametricLifetimeModel.moment.__doc__ = """
n-th order moment

Parameters
----------
n : order of the moment, at least 1.

Returns
-------
np.float64
    n-th order moment.
"""
FrozenParametricLifetimeModel.mean.__doc__ = MOMENT_BASE_DOCSTRING.format(name="The mean")
FrozenParametricLifetimeModel.var.__doc__ = MOMENT_BASE_DOCSTRING.format(name="The variance")
FrozenParametricLifetimeModel.median.__doc__ = MOMENT_BASE_DOCSTRING.format(name="The median")

FrozenParametricLifetimeModel.ichf.__doc__ = """
Inverse cumulative hazard function.

Parameters
----------
cumulative_hazard_rate : float or np.ndarray
    Cumulative hazard rate value(s) at which to compute the function.
    If ndarray, allowed shapes are ``()``, ``(n,)`` or ``(m, n)``.

Returns
-------
np.float64 or np.ndarray
    Function values at each given cumulative hazard rate(s).
"""

FrozenLifetimeRegression.dhf.__doc__ = TIME_BASE_DOCSTRING.format(name="The derivative of the hazard function")
FrozenLifetimeRegression.jac_hf.__doc__ = JAC_BASE_DOCSTRING.format(name="The jacobian of the hazard function")
FrozenLifetimeRegression.jac_chf.__doc__ = JAC_BASE_DOCSTRING.format(
    name="The jacobian of the cumulative hazard function"
)
FrozenLifetimeRegression.jac_sf.__doc__ = JAC_BASE_DOCSTRING.format(name="The jacobian of the survival function")
FrozenLifetimeRegression.jac_pdf.__doc__ = JAC_BASE_DOCSTRING.format(
    name="The jacobian of the probability density function"
)
FrozenLifetimeRegression.jac_cdf.__doc__ = JAC_BASE_DOCSTRING.format(
    name="The jacobian of the cumulative distribution function"
)



from abc import ABC
from collections.abc import Iterator
from typing import (
    Iterable,
    Optional,
    TypeVarTuple,
)

import numpy as np
from numpy.lib import recfunctions as rfn
from numpy.typing import NDArray
from typing_extensions import override

from relife.economic import ExponentialDiscounting, Reward
from relife.lifetime_model import (
    FrozenAgeReplacementModel,
    FrozenLeftTruncatedModel,
    FrozenLifetimeRegression,
    LifetimeDistribution,
)

from .renewal_process import RenewalProcess, RenewalRewardProcess

Args = TypeVarTuple("Args")


class CountDataIterator(Iterator[NDArray[np.void]], ABC):
    timeline: Optional[NDArray[np.float64]]
    start_counter: Optional[NDArray[np.uint32]]
    end_counter: Optional[NDArray[np.uint32]]

    def __init__(
        self,
        size: int | tuple[int] | tuple[int, int],
        window: tuple[float, float],
        seed: Optional[int] = None,
    ):
        self.size = size
        self.t0, self.tf = window
        self.seed = seed
        self.timeline = None
        self.stop_counter = None
        self.start_counter = None
        self.cycle = 0

    @property
    def stop(self) -> bool:
        if self.stop_counter is not None:
            return np.all(self.stop_counter > 0)
        return False

    @property
    def just_crossed_t0(self) -> Optional[NDArray[np.bool_]]:
        if self.start_counter is not None:
            return self.start_counter == 1
        return None

    @property
    def just_crossed_tf(self) -> Optional[NDArray[np.bool_]]:
        if self.stop_counter is not None:
            return self.stop_counter == 1
        return None

    @property
    def selection(self) -> Optional[NDArray[np.bool_]]:
        if self.start_counter is not None and self.stop_counter is not None:
            return np.logical_and(self.start_counter >= 1, self.stop_counter <= 1)
        return None

    def base_structarray(self) -> Optional[NDArray[np.void]]:
        selection = self.selection
        if selection is None:
            return None
        asset_id, sample_id = np.where(np.atleast_2d(selection))
        struct_array = np.zeros(
            sample_id.size,
            dtype=np.dtype(
                [
                    ("sample_id", np.uint32),  #  unsigned 32bit integer
                    ("asset_id", np.uint32),  #  unsigned 32bit integer
                ]
            ),
        )
        struct_array["sample_id"] = sample_id.astype(np.uint32)
        struct_array["asset_id"] = asset_id.astype(np.uint32)
        return struct_array


class RenewalProcessIterator(CountDataIterator):

    def __init__(
        self,
        process: RenewalProcess[
            LifetimeDistribution | FrozenLifetimeRegression | FrozenAgeReplacementModel | FrozenLeftTruncatedModel
        ],
        size: int | tuple[int] | tuple[int, int],
        window: tuple[float, float],
        seed: Optional[int] = None,
    ):
        super().__init__(size, window, seed=seed)
        self.process = process

    @property
    def model(
        self,
    ) -> LifetimeDistribution | FrozenLifetimeRegression | FrozenAgeReplacementModel | FrozenLeftTruncatedModel:
        if self.cycle == 0 and self.process.first_lifetime_model is not None:
            return self.process.first_lifetime_model
        return self.process.lifetime_model

    def step(self):
        # time is not residual age if return_entry is True (see LeftTruncatedModel)
        time, event, model_entry = self.model.rvs(size=self.size, return_event=True, return_entry=True, seed=self.seed)
        if self.cycle == 0:
            # initialize timeline, stop_counter and start_counter
            self.timeline = np.zeros_like(time, dtype=np.float64)  # ensure broadcasting
            self.stop_counter = np.zeros_like(self.timeline, dtype=np.uint32)
            self.start_counter = np.zeros_like(self.timeline, dtype=np.uint32)
        else:
            model_entry.fill(0.0)  # cancel any model entry after the first cycle

        # update timeline
        self.timeline += time

        # update start and stop counter
        self.start_counter[self.timeline > self.t0] += 1
        self.stop_counter[self.timeline > self.tf] += 1

        # tf right censorings
        # censored time : self.timeline - self.tf
        # observed time : time - censored time
        time = np.where(self.just_crossed_tf, time - (self.timeline - self.tf), time)
        self.timeline[self.just_crossed_tf] = self.tf
        event[self.just_crossed_tf] = False

        entry = np.zeros_like(model_entry)
        if self.cycle == 0:  # t0 added to model_entry
            entry = np.where(self.just_crossed_t0, self.t0 + model_entry, entry)
        else:
            # previous timeline step = self.timeline - time
            entry = np.where(self.just_crossed_t0, self.t0 - (self.timeline - time), entry)

        # update seed to avoid having the same rvs result
        if self.seed is not None:
            self.seed += 1
        # update cycle
        self.cycle += 1
        base_structarray = self.base_structarray()
        nb_renewal = np.full_like(self.timeline, self.cycle, dtype=np.uint32)

        struct_arr = rfn.append_fields(  #  works on structured_array too
            base_structarray,  # add to struct of timeline, sample_id, asset_id
            ("timeline", "time", "event", "entry", "nb_renewal"),
            (
                self.timeline[self.selection],
                time[self.selection],
                event[self.selection],
                entry[self.selection],
                nb_renewal[self.selection],
            ),
            (np.float64, np.float64, np.bool_, np.float64, np.uint32),
            usemask=False,
            asrecarray=False,
        )
        return struct_arr

    def __next__(self) -> NDArray[np.void]:
        while not self.stop:
            struct_arr = self.step()
            while struct_arr.size == 0 and not self.stop:  # skip cycles while arrays are empty (if t0 != 0.)
                struct_arr = self.step()
            return struct_arr
        raise StopIteration


class RenewalRewardProcessIterator(RenewalProcessIterator):
    reward: Reward
    discounting: ExponentialDiscounting

    def __init__(
        self,
        process: RenewalRewardProcess[
            LifetimeDistribution | FrozenLifetimeRegression | FrozenAgeReplacementModel | FrozenLeftTruncatedModel,
            Reward,
        ],
        size: int | tuple[int] | tuple[int, int],
        window: tuple[float, float],
        seed: Optional[int] = None,
    ):
        super().__init__(process, size, window, seed)
        self.reward = self.process.reward
        self.discounting = self.process.discounting

    @override
    def __next__(self) -> NDArray[np.void]:
        struct_array = super().__next__()
        # may be type hint error in rfn.append_fields overload
        return rfn.append_fields(
            struct_array,
            "reward",
            self.reward.sample(struct_array["time"]) * self.discounting.factor(struct_array["timeline"]),
            np.float64,
            usemask=False,
            asrecarray=False,
        )  # type: ignore


class RenewalProcessIterable(Iterable):

    def __init__(
        self,
        process: RenewalProcess[
            LifetimeDistribution | FrozenLifetimeRegression | FrozenAgeReplacementModel | FrozenLeftTruncatedModel
        ],
        size: int | tuple[int] | tuple[int, int],
        window: tuple[float, float],
        seed: Optional[int] = None,
    ):
        self.process = process
        self.size = size
        self.window = window
        self.seed = seed

    def __iter__(self) -> RenewalProcessIterator:
        if isinstance(self.process, RenewalProcess):
            return RenewalProcessIterator(self.process, self.size, self.window, self.seed)
        else:
            return RenewalRewardProcessIterator(self.process, self.size, self.window, self.seed)


# class NonHomogeneousPoissonProcessIterator(SampleIterator):
#
#     rewards: Optional[Reward]
#     discounting: Optional[Discounting]
#     hpp_timeline: Optional[NDArray[np.float64]]
#     failure_times: Optional[NDArray[np.float64]]
#     ages: Optional[NDArray[np.float64]]
#     ar: Optional[NDArray[np.float64]]
#     is_new_asset: Optional[NDArray[np.bool_]]
#     entries: Optional[NDArray[np.float64]]
#     renewals_ids: Optional[NDArray[np.int64]]
#
#     def __init__(
#         self,
#         size: int,
#         tf: float,  # calendar end time
#         t0: float = 0.0,  # calendar beginning time
#         *,
#         seed: Optional[int] = None,
#         keep_last: bool = True,
#     ):
#         super().__init__(size, tf, t0, seed=seed, keep_last=keep_last)
#
#         self.rewards = None
#         self.discounting = None
#
#         self.hpp_timeline = None  # exposed attribute (set/get)
#         self.failure_times = None
#         self.ages = None
#         # self._assets_ids = None
#         self.ar = None
#         self.is_new_asset = None
#         self.entries = None
#         self.renewals_ids = None
#         self.exponential_dist = Exponential(1.0)
#
#     def compute_rewards(
#         self, timeline: NDArray[np.float64], ages: NDArray[np.float64]
#     ) -> NDArray[np.float64]:
#         rewards = np.zeros_like(ages)
#         if self.rewards and self.discounting:
#             rewards = self.rewards(ages) * self.discounting.factor(timeline)
#         if self.rewards and not self.discounting:
#             rewards = self.rewards(ages)
#         return rewards
#
#     def set_model(
#         self,
#         model: FrozenParametricLifetimeModel,
#         ar: Optional[NDArray[np.float64]] = None,
#     ) -> None:
#
#         if self.model is None:
#             # self._nb_assets = get_nb_assets(model_args)
#             self.timeline = np.zeros((model.args_nb_assets, self.size))
#             # counting arrays to catch values crossing t0 and tf bounds
#             self.stop_counter = np.zeros((model.args_nb_assets, self.size), dtype=np.int64)
#             self.start_counter = np.zeros((model.args_nb_assets, self.size), dtype=np.int64)
#
#             self.hpp_timeline = np.zeros((model.args_nb_assets, self.size))
#             self.failure_times = np.zeros((model.args_nb_assets, self.size))
#             self.ages = np.zeros((model.args_nb_assets, self.size))
#             self.entries = np.zeros((model.args_nb_assets, self.size))
#             self.is_new_asset = np.zeros((model.args_nb_assets, self.size), dtype=np.bool_)
#             self.renewals_ids = np.zeros((model.args_nb_assets, self.size), dtype=np.int64)
#
#         self.model = model
#         self.ar = (
#             ar if ar is not None else (np.ones(self.model) * np.inf).reshape(-1, 1)
#         )
#
#     def step(self) -> dict[str, AnyNDArray]:
#
#         # reset those who are replaced
#         self.ages[self.is_new_asset] = 0.0  # asset is replaced (0 aged asset)
#         # self._assets_ids[self.is_new_asset] += 1 # asset is replaced (new asset id)
#         self.hpp_timeline[self.is_new_asset] = 0.0  # reset timeline
#         self.failure_times[self.is_new_asset] = 0.0
#         self.entries[self.is_new_asset] = 0.0
#         self.renewals_ids[self.is_new_asset] += 1
#         self.is_new_asset.fill(False)  # reset to False
#
#         # generate new values
#         self.hpp_timeline += self.exponential_dist.rvs(
#             size=self.size * self.model.args_nb_assets, seed=self.seed
#         ).reshape((self.model.args_nb_assets, self.size))
#
#         failure_times = self.model.ichf(self.hpp_timeline)
#         durations = failure_times - self.failure_times  # t_i+1 - t_i
#         self.failure_times = failure_times.copy()  # update t_i <- t_i+1
#         self.timeline += durations
#         self.ages += durations
#
#         # create array of events_indicators
#         events_indicators = np.ones_like(self.ages, np.bool_)
#
#         # ar update (before because it changes timeline, thus start and stop conditions)
#         self.timeline = np.where(
#             self.ages >= self.ar,
#             self.timeline - (self.ages - np.ones_like(self.timeline) * self.ar),
#             self.timeline,
#         )  # substract time after ar
#         self.ages = np.where(
#             self.ages >= self.ar, np.ones_like(self.ages) * self.ar, self.ages
#         )  # set ages to ar
#         self.is_new_asset[
#             np.logical_and(self.ages >= self.ar, ~self.just_crossed_tf)
#         ] = True
#         events_indicators[
#             np.logical_and(self.ages >= self.ar, ~self.just_crossed_tf)
#         ] = False
#
#         # update stop conditions
#         self.start_counter[self.timeline > self.t0] += 1
#         self.stop_counter[self.timeline > self.tf] += 1
#
#         # t0 entries update
#         self.entries = np.where(
#             self.just_crossed_t0, self.ages - (self.timeline - self.t0), self.entries
#         )
#
#         # tf update censoring update
#         self.ages = np.where(
#             self.just_crossed_tf, self.ages - (self.timeline - self.tf), self.ages
#         )
#         self.timeline[self.just_crossed_tf] = self.tf
#         events_indicators[self.just_crossed_tf] = False
#
#         # returned entries
#         entries = self.entries.copy()
#         # update for next iteration (keep previous ages)
#         self.entries = np.where(
#             events_indicators, self.ages, self.entries
#         )  # keep previous ages as entry for next iteration
#
#         # update seed to avoid having the same rvs result
#         if self.seed is not None:
#             self.seed += 1
#
#         rewards = self.compute_rewards(self.timeline, self.ages)
#
#         return self.construct_sample_data1d(
#             ages=self.ages,
#             renewals_ids=self.renewals_ids,
#             entries=entries,
#             events_indicators=events_indicators,
#             rewards=rewards,
#         )


#
# @sample_count_data.register
# def _(
#     obj: Union[
#         NonHomogeneousPoissonProcess,
#         NonHomogeneousPoissonAgeReplacementPolicy,
#     ],
#     size: int,
#     tf: float,
#     t0: float = 0.0,
#     maxsample: int = 1e5,
#     seed: Optional[int] = None,
# ):
#     keys = (
#         "timeline",
#         "ages",
#         "events_indicators",
#         "samples_ids",
#         "assets_ids",
#         "rewards",
#     )
#     iterator = NonHomogeneousPoissonProcessIterator(size, tf, t0=t0, seed=seed)
#     iterator.set_model(obj.model, ar=getattr(obj, "ar", None))
#     if isinstance(
#         obj,
#         NonHomogeneousPoissonAgeReplacementPolicy,
#     ):
#         iterator.rewards = age_replacement_rewards(obj.ar, obj.cr, obj.cp)
#         iterator.discounting = obj.discounting
#     stack = stack1d(iterator, keys, maxsample=maxsample)
#
#     return NHPPCountData(t0, tf, **stack)


# @failure_data_sample.register
# def _(
#     obj: Union[
#         NonHomogeneousPoissonProcess,
#         NonHomogeneousPoissonAgeReplacementPolicy,
#     ],
#     size: int,
#     tf: float,
#     t0: float = 0.0,
#     maxsample: int = 1e5,
#     seed: Optional[int] = None,
#     use: str = "model",
# ):
#     keys = (
#         "timeline",
#         "samples_ids",
#         "assets_ids",
#         "ages",
#         "entries",
#         "events_indicators",
#         "renewals_ids",
#     )
#
#     if use != "model":
#         raise ValueError("Invalid 'use' value. Only 'model' can be set")
#
#     iterator = NonHomogeneousPoissonProcessIterator(size, tf, t0=t0, seed=seed, keep_last=True)
#     iterator.set_model(obj.model, ar=obj.ar if hasattr(obj, "ar") else None)
#
#     stack = stack1d(iterator, keys, maxsample=maxsample)
#
#     str_samples_ids = np.char.add(
#         np.full_like(stack["samples_ids"], "S", dtype=np.str_),
#         stack["samples_ids"].astype(np.str_),
#     )
#     str_assets_ids = np.char.add(
#         np.full_like(stack["assets_ids"], "A", dtype=np.str_),
#         stack["assets_ids"].astype(np.str_),
#     )
#     str_renewals_ids = np.char.add(
#         np.full_like(stack["assets_ids"], "R", dtype=np.str_),
#         stack["renewals_ids"].astype(np.str_),
#     )
#     assets_ids = np.char.add(str_samples_ids, str_assets_ids)
#     assets_ids = np.char.add(assets_ids, str_renewals_ids)
#
#     sort_ind = np.lexsort((stack["timeline"], assets_ids))
#
#     entries = stack["entries"][sort_ind]
#     events_indicators = stack["events_indicators"][sort_ind]
#     ages = stack["ages"][sort_ind]
#     assets_ids = assets_ids[sort_ind]
#
#     # print("assets_ids", assets_ids)
#     # print("timeline", timeline)
#     # print("ages", ages)
#     # print("events_indicators", events_indicators)
#     # print("entries", entries)
#
#     first_ages_index = np.roll(assets_ids, 1) != assets_ids
#     last_ages_index = np.roll(first_ages_index, -1)
#
#     immediatly_replaced = np.logical_and(~events_indicators, first_ages_index)
#
#     # print("first_ages_index", first_ages_index)
#     # print("last_ages_index", last_ages_index)
#     # print("immediatly_replaced", immediatly_replaced)
#
#     # prefix = np.full_like(assets_ids[immediatly_replaced], "Z", dtype=np.str_)
#     # _assets_ids = np.char.add(prefix, assets_ids[immediatly_replaced])
#     _assets_ids = assets_ids[immediatly_replaced]
#     first_ages = entries[immediatly_replaced].copy()
#     last_ages = ages[immediatly_replaced].copy()
#
#     # print("assets_ids", _assets_ids)
#     # print("first_ages", first_ages)
#     # print("last_ages", last_ages)
#
#     events_assets_ids = assets_ids[events_indicators]
#     events_ages = ages[events_indicators]
#     other_assets_ids = np.unique(events_assets_ids)
#     _assets_ids = np.concatenate((_assets_ids, other_assets_ids))
#     first_ages = np.concatenate(
#         (first_ages, entries[first_ages_index & events_indicators])
#     )
#     last_ages = np.concatenate(
#         (last_ages, ages[last_ages_index & ~immediatly_replaced])
#     )
#
#     # print("events_assets_ids", events_assets_ids)
#     # print("events_ages", events_ages)
#     # print("assets_ids", _assets_ids)
#     # print("first_ages", first_ages)
#     # print("last_ages", last_ages)
#
#     # last sort (optional but convenient to control data)
#     sort_ind = np.argsort(events_assets_ids)
#     events_assets_ids = events_assets_ids[sort_ind]
#
#     sort_ind = np.argsort(_assets_ids)
#     _assets_ids = _assets_ids[sort_ind]
#     first_ages = first_ages[sort_ind]
#     last_ages = last_ages[sort_ind]
#
#     return events_assets_ids, events_ages, _assets_ids, first_ages, last_ages



from __future__ import annotations

from itertools import chain
from typing import TYPE_CHECKING, Any, Optional, Self, Union, overload

import numpy as np
from numpy.typing import NDArray

if TYPE_CHECKING:
    from relife.lifetime_model import (
        AgeReplacementModel,
        FrozenAgeReplacementModel,
        FrozenLeftTruncatedModel,
        FrozenLifetimeRegression,
        FrozenParametricLifetimeModel,
        LeftTruncatedModel,
        LifetimeDistribution,
        LifetimeRegression,
        ParametricLifetimeModel,
    )
    from relife.stochastic_process import (
        FrozenNonHomogeneousPoissonProcess,
        NonHomogeneousPoissonProcess,
    )


class ParametricModel:
    """
    Base class to create a parametric_model core.

    Any parametric_model core must inherit from `ParametricModel`.
    """

    _nested_models: dict[str, ParametricModel]

    def __init__(self, **kwparams: Optional[float]):
        self._params = Parameters(**kwparams)
        self._nested_models = {}

    @property
    def params(self) -> NDArray[np.float64]:
        """
        Parameters values.

        Returns
        -------
        ndarray
            Parameters values of the core

        Notes
        -----
        If parameter values are not set, they are encoded as `np.nan` value.

        Parameters can be by manually setting`params` through its setter, fitting the core if `fit` exists or
        by specifying all parameters values when the core object is initialized.
        """
        return np.array(self._params.get_values())

    @params.setter
    def params(self, new_params: NDArray[np.float64]):
        if not isinstance(new_params, np.ndarray):
            raise ValueError(
                f"Incorrect params values. It must be contained in a 1d array. Got type {type(new_params)}"
            )
        if new_params.ndim > 1:
            raise ValueError(f"Expected params values to be 1d array. Got {new_params.ndim} ndim")
        self._params.set_values(tuple(v.item() for v in new_params))

    @property
    def params_names(self) -> tuple[str, ...]:
        """
        Parameters names.

        Returns
        -------
        list of str
            Parameters names

        Notes
        -----
        Parameters values can be requested (a.k.a. get) by their name at instance level.
        """
        return self._params.get_names()

    @property
    def nb_params(self) -> int:
        """
        Number of parameters.

        Returns
        -------
        int
            Number of parameters.

        """
        return self._params.size

    def _set_nested_model(self, name: str, model: ParametricModel):
        """Compose with new ``ParametricModel`` instance(s).

        This method acts like function. When you set a nested model instance, the followings happen :
        - each new parameters are added to the current ``Parameters`` instance
        - each new `ParametricModel` instance is accessible as a standard attribute

        Parameters
        ----------
        name : str
            model name
        model : ParametricModel
            model instance
        """
        self._nested_models[name] = model
        self._params.set_leaf(f"{name}.params", getattr(model, "_params"))

    def __getattr__(self, name: str):
        class_name = type(self).__name__
        if name in self.__dict__:
            return self.__dict__[name]
        if name in super().__getattribute__("_params").get_names():
            return super().__getattribute__("_params").get_param_value(name)
        if name in super().__getattribute__("_nested_models"):
            return super().__getattribute__("_nested_models").get(name)
        raise AttributeError(f"{class_name} has no attribute named {name}")

    def __setattr__(self, name: str, value: Any):
        if name in ("_params", "_nested_models"):
            super().__setattr__(name, value)
        elif name in self._params.get_names():
            self._params.set_param_value(name, value)
        elif isinstance(value, ParametricModel):
            self._set_nested_model(name, value)
        else:  # just set a new attribute
            super().__setattr__(name, value)


# MAYBE, custom array container can replace it : https://numpy.org/doc/stable/user/basics.dispatch.html#writing-custom-array-containers
class Parameters:
    """
    Dict-like tree structured parameters.

    Every ``ParametricModel`` are composed of a ``Parameters`` instance.
    """

    _parent: Optional[Self]
    _leaves: dict[str, Self]
    _nodemapping: dict[str, float]  # np.nan is float
    _names: tuple[str, ...]
    _values: tuple[float, ...]

    def __init__(self, **kwargs: Optional[float]):
        self._parent = None
        self._leaves = {}
        self._nodemapping = {}
        self._values = ()
        self._names = ()
        if bool(kwargs):
            self.set_node(kwargs)

    def get_names(self) -> tuple[str, ...]:
        return self._names

    def get_values(self) -> tuple[float, ...]:
        return self._values

    def get_leaf(self, name: str) -> Parameters:
        try:
            return self._leaves[name]
        except KeyError:
            raise ValueError(f"Parameters object does not have leaf parameters called {name} in its scope")

    def get_param_value(self, name: str) -> float:
        try:
            return self._nodemapping[name]
        except KeyError:
            raise ValueError(f"Parameters object does not have parameter name called {name} in its scope")

    def set_param_value(self, name: str, value: float) -> None:
        if name not in self._nodemapping:
            raise ValueError(f"Parameters object does not have parameter name called {name} in its scope")
        self._nodemapping[name] = value
        self._update_names_and_values()

    @property
    def size(self) -> int:
        return len(self._values)

    def set_node(self, mapping: dict[str, Optional[float]]) -> None:
        """
        set node dict
        """
        self._nodemapping = {k: v if v is not None else np.nan for k, v in mapping.items()}
        self._update_names_and_values()  # update _names and _values

    def set_leaf(self, leaf_name: str, leaf: Parameters) -> None:
        """
        set a leaf or new leaf
        """
        if leaf_name not in self._leaves:
            leaf._parent = self
        self._leaves[leaf_name] = leaf
        self._update_names_and_values()  # update _names and _values

    def set_values(self, values: tuple[float, ...]) -> None:
        """set values of all tree"""
        if len(values) != self.size:
            raise ValueError(f"Expected {self.size} values but got {len(values)}")
        pos = len(self._nodemapping)
        self._nodemapping.update(zip(self._nodemapping.keys(), (np.nan if v is None else v for v in values[:pos])))
        self._values = tuple((np.nan if v is None else v for v in values))
        for leaf in self._leaves.values():
            leaf.set_values(values[pos : pos + leaf.size])
            pos += leaf.size
        if self._parent is not None:
            self._parent._update_names_and_values()

    def _update_names_and_values(self) -> None:
        """update names and values of current and parent nodes"""

        def items_walk(parameters: Self):
            yield tuple(parameters._nodemapping.items())
            for leaf in parameters._leaves.values():
                yield tuple(chain.from_iterable(items_walk(leaf)))

        generator = chain.from_iterable(items_walk(self))
        try:
            self._names, self._values = zip(*generator)
        except StopIteration:
            pass
        # recursively update parent
        if self._parent is not None:
            self._parent._update_names_and_values()


def model_args_names(
    model: Union[
        LifetimeDistribution, LifetimeRegression, LeftTruncatedModel, AgeReplacementModel, NonHomogeneousPoissonProcess
    ],
) -> tuple[str, ...]:
    from relife.lifetime_model import (
        AgeReplacementModel,
        FrozenParametricLifetimeModel,
        LeftTruncatedModel,
        LifetimeDistribution,
        LifetimeRegression,
    )
    from relife.stochastic_process import (
        NonHomogeneousPoissonProcess,
    )

    args_names = ()
    if isinstance(model, LifetimeRegression):
        args_names = args_names + ("covar",) + model_args_names(model.baseline)
    elif isinstance(model, AgeReplacementModel):
        args_names = args_names + ("ar",) + model_args_names(model.baseline)
    elif isinstance(model, LeftTruncatedModel):
        args_names = args_names + ("a0",) + model_args_names(model.baseline)
    elif isinstance(model, NonHomogeneousPoissonProcess):
        args_names = model_args_names(model.baseline)
    elif isinstance(model, (LifetimeDistribution, FrozenParametricLifetimeModel)):
        return args_names
    else:
        raise ValueError
    return args_names


@overload
def freeze(model: LifetimeRegression, **kwargs: float | NDArray[np.float64]) -> FrozenLifetimeRegression: ...


@overload
def freeze(model: LeftTruncatedModel, **kwargs: float | NDArray[np.float64]) -> FrozenLeftTruncatedModel: ...


@overload
def freeze(model: AgeReplacementModel, **kwargs: float | NDArray[np.float64]) -> FrozenAgeReplacementModel: ...


@overload
def freeze(
    model: NonHomogeneousPoissonProcess, **kwargs: float | NDArray[np.float64]
) -> FrozenNonHomogeneousPoissonProcess: ...


def freeze(
    model: LifetimeRegression | LeftTruncatedModel | AgeReplacementModel | NonHomogeneousPoissonProcess,
    **kwargs: float | NDArray[np.float64],
) -> (
    FrozenLifetimeRegression | FrozenLeftTruncatedModel | FrozenAgeReplacementModel | FrozenNonHomogeneousPoissonProcess
):
    """
    Freeze a parametric model with given arguments.

    This function takes a parametric model instance (lifetime model or stochastic process) and stores the given arguments
    in a new model interface allowing to resquest functions only with ``time`` argument.

    Parameters
    ----------
    model: LifetimeRegression, LeftTruncatedModel, AgeReplacementModel or NonHomogeneousPoissonProcess
        The model to be frozen. It should be one of the supported model types.
    **kwargs: float or np.ndarray
        Keyword arguments representing the parameter names and their respective values to be frozen.

    Returns
    -------
    FrozenLifetimeRegression, FrozenLeftTruncatedModel, FrozenAgeReplacementModel or FrozenNonHomogeneousPoissonProcess
        The frozen version of the input model with immutable provided argument values.

    Raises
    ------
    ValueError
        Raised when the provided arguments do not match the expected argument names
        for the given model, or if the model is not one of the supported types.


    Examples
    --------
    >>> from relife import freeze
    >>> from relife.lifetime_model import Weibull, ProportionalHazard
    >>> weibull = Weibull(3.5, 0.01)
    >>> regression = ProportionalHazard(weibull, coefficients=(1., 2.))
    >>> frozen_regression = freeze(regression, covar=1.5)
    >>> sf = frozen_regression.sf(np.array([10, 20])) # covar is fixed at 1.5


    """

    from relife.lifetime_model import (
        AgeReplacementModel,
        FrozenAgeReplacementModel,
        FrozenLeftTruncatedModel,
        FrozenLifetimeRegression,
        LeftTruncatedModel,
        LifetimeRegression,
    )
    from relife.stochastic_process import (
        FrozenNonHomogeneousPoissonProcess,
        NonHomogeneousPoissonProcess,
    )

    args_names = model_args_names(model)
    set_args_names = set(args_names)
    set_keys = set(kwargs.keys())
    if set_args_names != set_keys:
        raise ValueError(f"Expected {set_args_names} but got {set_keys}")

    position_mapping = {name: i for i, name in enumerate(args_names)}
    reorder_kwargs = dict(sorted(kwargs.items(), key=lambda item: position_mapping[item[0]]))
    args_values: tuple[float | NDArray[np.float64], ...] = tuple(reorder_kwargs.values())

    frozen_model: Union[
        FrozenLifetimeRegression,
        FrozenLeftTruncatedModel,
        FrozenAgeReplacementModel,
        FrozenNonHomogeneousPoissonProcess,
    ]
    # here args_nb_assets is set to 1 by default and then, will be overriden after testing frozen_args coherence
    if isinstance(model, LifetimeRegression):
        frozen_model = FrozenLifetimeRegression(model, 1, args_values[0], *args_values[1:])
    elif isinstance(model, AgeReplacementModel):
        frozen_model = FrozenAgeReplacementModel(model, 1, args_values[0], *args_values[1:])
    elif isinstance(model, LeftTruncatedModel):
        frozen_model = FrozenLeftTruncatedModel(model, 1, args_values[0], *args_values[1:])
    elif isinstance(model, NonHomogeneousPoissonProcess):
        frozen_model = FrozenNonHomogeneousPoissonProcess(model, 1, *args_values)
    else:
        raise ValueError
    frozen_model.args_nb_assets = get_args_nb_assets(frozen_model)
    return frozen_model


def get_args_nb_assets(
    model: Union[
        FrozenLifetimeRegression,
        FrozenLeftTruncatedModel,
        FrozenAgeReplacementModel,
        FrozenNonHomogeneousPoissonProcess,
    ],
) -> int:

    from relife.lifetime_model import (
        AgeReplacementModel,
        FrozenAgeReplacementModel,
        FrozenLeftTruncatedModel,
        FrozenLifetimeRegression,
        LeftTruncatedModel,
        LifetimeRegression,
    )
    from relife.stochastic_process import FrozenNonHomogeneousPoissonProcess

    local_model: (
        ParametricLifetimeModel[*tuple[float | NDArray[np.float64], ...]]
        | FrozenParametricLifetimeModel[*tuple[float | NDArray[np.float64], ...]]
    )
    arg_nb_assets: int = 1
    arg_name: Optional[str] = None
    if isinstance(model, FrozenNonHomogeneousPoissonProcess):
        local_model = model.unfreeze().baseline
    else:
        local_model = model

    _arg_nb_assets: int = 1
    for arg in model.frozen_args:
        if isinstance(local_model, FrozenLifetimeRegression):
            # arg is covar
            _arg_nb_assets = np.atleast_2d(np.asarray(arg)).shape[0]
            arg_name = "covar"
            local_model = local_model.unfreeze().baseline
        elif isinstance(local_model, LifetimeRegression):
            # arg is covar
            _arg_nb_assets = np.atleast_2d(np.asarray(arg)).shape[0]
            arg_name = "covar"
            local_model = local_model.baseline
        elif isinstance(local_model, FrozenLeftTruncatedModel):
            # arg is a0
            _arg_nb_assets = np.asarray(arg).size
            arg_name = "a0"
            local_model = local_model.unfreeze().baseline
        elif isinstance(local_model, LeftTruncatedModel):
            # arg is a0
            _arg_nb_assets = np.asarray(arg).size
            arg_name = "a0"
            local_model = local_model.baseline
        elif isinstance(local_model, FrozenAgeReplacementModel):
            # arg is ar
            _arg_nb_assets = np.asarray(arg).size
            arg_name = "ar"
            local_model = local_model.unfreeze().baseline
        elif isinstance(local_model, AgeReplacementModel):
            # arg is ar
            _arg_nb_assets = np.asarray(arg).size
            arg_name = "ar"
            local_model = local_model.baseline
        if arg_name is not None:
            # test if nb_assets changed and would not be broadcastable
            if _arg_nb_assets != 1 and arg_nb_assets != 1 and _arg_nb_assets != arg_nb_assets:
                raise ValueError(
                    f"Invalid number of assets given in arguments. Got several nb assets. {arg_name} has {_arg_nb_assets} but already got {arg_nb_assets}"
                )
            arg_nb_assets = _arg_nb_assets
    return arg_nb_assets


# see sklearn/base.py : return unfitted ParametricModel
# def clone(model: ParametricModel) -> ParametricModel: ...
